package dev.fitko.fitconnect.client;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.config.Environment;
import dev.fitko.fitconnect.api.config.EnvironmentName;
import dev.fitko.fitconnect.api.config.ResourcePaths;
import dev.fitko.fitconnect.api.config.SenderConfig;
import dev.fitko.fitconnect.api.config.SubscriberConfig;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.PublicServiceType;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.client.factory.ClientFactory;
import dev.fitko.fitconnect.client.model.AttachmentPayload;
import dev.fitko.fitconnect.client.model.DataPayload;
import dev.fitko.fitconnect.client.model.ReceivedSubmission;
import dev.fitko.fitconnect.client.model.ServiceTypePayload;
import dev.fitko.fitconnect.client.util.SubmissionUtil;
import dev.fitko.fitconnect.core.crypto.HashService;
import dev.fitko.fitconnect.core.crypto.JWECryptoService;
import org.apache.tika.mime.MimeTypes;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.condition.EnabledIfEnvironmentVariable;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.text.ParseException;
import java.util.*;
import java.util.stream.Collectors;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.hasItems;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertTrue;

/**
 * This test uses real credentials and endpoints stored in gitlab ci variables to test the sdk against a real system
 */
class ClientIntegrationTest {

    final String authBaseUrl = "https://auth-testing.fit-connect.fitko.dev";
    final String submissionBaseUrl = "https://submission-api-testing.fit-connect.fitko.dev";

    @Test
    @EnabledIfEnvironmentVariable(named = "TEST_DESTINATION_ID", matches = ".*")
    void testSendAndConfirmCycle() {

        // Given
        final var testEnv = new Environment(authBaseUrl, submissionBaseUrl, true);
        final ApplicationConfig config = getConfigWithCredentialsFromGitlab(testEnv);

        final var sentSubmission = ClientFactory.senderClient(config)
                .newSubmission()
                .withAttachment(new File("src/test/resources/attachment.txt"))
                .withJsonData("{ data: 'Beispiel Fachdaten' }")
                .withDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID")))
                .withServiceType("Test Service", "urn:de:fim:leika:leistung:99400048079000")
                .submit();

        assertTrue(sentSubmission.isPresent());

        // When
        final Optional<ReceivedSubmission> receivedSubmission =
                ClientFactory.subscriberClient(config)
                        .requestSubmission(sentSubmission.get().getSubmissionId());

        // Then
        assertTrue(receivedSubmission.isPresent());
        assertThat(receivedSubmission.get().getData(), is("{ data: 'Beispiel Fachdaten' }"));
        assertThat(receivedSubmission.get().getMimeType().toString(), is("application/json"));
        assertThat(new String(receivedSubmission.get().getAttachments().get(0).getDecryptedData()), is("Test attachment"));
    }

    @Test
    @EnabledIfEnvironmentVariable(named = "TEST_DESTINATION_ID", matches = ".*")
    void testSendAndConfirmCycleWithEncryptedData() throws ParseException, IOException {

        // Given
        final UUID destinationId = UUID.fromString(System.getenv("TEST_DESTINATION_ID"));
        final var testEnv = new Environment(authBaseUrl, submissionBaseUrl, true);
        final ApplicationConfig config = getConfigWithCredentialsFromGitlab(testEnv);
        final CryptoService cryptoService = new JWECryptoService(new HashService());
        final SenderClient.Start client = ClientFactory.senderClient(config);

        final Optional<String> publicKey = client.getPublicKey(destinationId);
        assertTrue(publicKey.isPresent());

        final RSAKey encryptionKey = RSAKey.parse(publicKey.get());

        final String jsonData = "{ data: 'Beispiel Fachdaten' }";
        final File attachmentFile = new File("src/test/resources/attachment.txt");
        final String attachmentData = Files.readString(attachmentFile.toPath());

        final String encryptedData = cryptoService.encryptString(encryptionKey, jsonData);
        final String encryptedAttachment = cryptoService.encryptString(encryptionKey, attachmentData);

        final DataPayload dataPayload = DataPayload.builder()
                .encryptedData(encryptedData)
                .mimeType(MimeType.APPLICATION_JSON)
                .schemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                .hashedData(cryptoService.hashBytes(jsonData.getBytes(StandardCharsets.UTF_8)))
                .build();

        final AttachmentPayload attachmentPayload = AttachmentPayload.builder()
                .encryptedData(encryptedAttachment)
                .file(attachmentFile)
                .mimeType(MimeTypes.PLAIN_TEXT)
                .attachmentId(UUID.randomUUID())
                .hashedData(cryptoService.hashBytes(attachmentData.getBytes(StandardCharsets.UTF_8)))
                .build();

        final ServiceTypePayload serviceTypePayload = ServiceTypePayload.builder()
                .name("Test Service")
                .leikaKey("urn:de:fim:leika:leistung:99400048079000")
                .build();

        final Data data = SubmissionUtil.buildData(dataPayload);
        final PublicServiceType publicServiceType = SubmissionUtil.buildPublicServiceType(serviceTypePayload);
        final List<Attachment> attachmentMetadata = SubmissionUtil.toAttachmentMetadata(List.of(attachmentPayload));
        final Metadata metadata = SubmissionUtil.buildMetadata(attachmentMetadata, data, publicServiceType);

        final String encryptedMetadata = cryptoService.encryptBytes(encryptionKey, new ObjectMapper().writeValueAsBytes(metadata));

        // When
        final var sentSubmission = client
                .newSubmissionWithEncryptedData()
                .withEncryptedAttachment(attachmentPayload.getAttachmentId(), attachmentPayload.getEncryptedData())
                .withEncryptedJsonData(encryptedData)
                .withEncryptedMetadata(encryptedMetadata)
                .withDestination(destinationId)
                .withServiceType("Test Service", "urn:de:fim:leika:leistung:99400048079000")
                .submit();

        assertTrue(sentSubmission.isPresent());

        final Optional<ReceivedSubmission> receivedSubmission =
                ClientFactory.subscriberClient(config)
                        .requestSubmission(sentSubmission.get().getSubmissionId());

        // Then
        assertTrue(receivedSubmission.isPresent());
        assertThat(receivedSubmission.get().getData(), is(jsonData));
        assertThat(receivedSubmission.get().getMimeType().toString(), is("application/json"));
        assertThat(new String(receivedSubmission.get().getAttachments().get(0).getDecryptedData()), is("Test attachment"));
    }

    @Test
    @EnabledIfEnvironmentVariable(named = "TEST_DESTINATION_ID", matches = ".*")
    void testListSubmissions() {

        // Given
        final var testEnv = new Environment(authBaseUrl, submissionBaseUrl, true);
        final ApplicationConfig config = getConfigWithCredentialsFromGitlab(testEnv);

        final var senderClient = ClientFactory.senderClient(config);
        final var subscriberClient = ClientFactory.subscriberClient(config);

        final UUID destinationId = UUID.fromString(System.getenv("TEST_DESTINATION_ID"));
        final String leikaKey = "urn:de:fim:leika:leistung:99400048079000";
        final String serviceName = "Test Service";

        final var submissionOne = senderClient.newSubmission()
                .withJsonData("{ data: 'Beispiel Fachdaten 1' }")
                .withDestination(destinationId)
                .withServiceType(serviceName, leikaKey)
                .submit();

        final var submissionTwo = senderClient.newSubmission()
                .withJsonData("{ data: 'Beispiel Fachdaten 2' }")
                .withDestination(destinationId)
                .withServiceType(serviceName, leikaKey)
                .submit();

        assertTrue(submissionOne.isPresent());
        assertTrue(submissionTwo.isPresent());

        // When
        final Set<SubmissionForPickup> submissions = subscriberClient.getAvailableSubmissions(destinationId);

        // Then
        assertFalse(submissions.isEmpty());

        final List<UUID> submissionIds = submissions.stream().map(SubmissionForPickup::getSubmissionId).collect(Collectors.toList());

        assertThat(submissionIds, hasItems(submissionOne.get().getSubmissionId(), submissionTwo.get().getSubmissionId()));

        // remove by confirming
        subscriberClient.requestSubmission(submissionOne.get().getSubmissionId());
        subscriberClient.requestSubmission(submissionTwo.get().getSubmissionId());
    }

    @Test
    @EnabledIfEnvironmentVariable(named = "TEST_DESTINATION_ID", matches = ".*")
    void testAbortedSendSubmissionWithKeyValidationNotSilent() {

        // Given
        final var prodEnv = new Environment(authBaseUrl, submissionBaseUrl, false);
        final ApplicationConfig config = getConfigWithCredentialsFromGitlab("PROD", prodEnv);

        // When
        final var sentSubmission = ClientFactory.senderClient(config).newSubmission()
                .withAttachment(new File("src/test/resources/attachment.txt"))
                .withJsonData("{ data: 'Beispiel Fachdaten' }")
                .withDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID")))
                .withServiceType("Test Service", "urn:de:fim:leika:leistung:99400048079000")
                .submit();

        // Then
        assertTrue(sentSubmission.isEmpty());
    }

    private ApplicationConfig getConfigWithCredentialsFromGitlab(final Environment env) {
        return getConfigWithCredentialsFromGitlab("TESTING", env);
    }

    private ApplicationConfig getConfigWithCredentialsFromGitlab(final String environmentName, final Environment env) {

        final var sender = new SenderConfig(System.getenv("SENDER_CLIENT_ID"), System.getenv("SENDER_CLIENT_SECRET"));

        final var subscriber = SubscriberConfig.builder()
                .clientId(System.getenv("SUBSCRIBER_CLIENT_ID"))
                .clientSecret(System.getenv("SUBSCRIBER_CLIENT_SECRET"))
                .privateDecryptionKeyPath("src/test/resources/private_decryption_test_key.json")
                .build();

        final var resourcePaths = ResourcePaths.builder()
                .authTokenPath("/token")
                .destinationPath("/v1/destinations/{destinationId}")
                .destinationKeyPath("/v1/destinations/{destinationId}/keys/{kid}")
                .eventLogPath("/v1/cases/{caseId}/events")
                .submissionsPath("/v1/submissions")
                .submissionPath("/v1/submissions/{submissionId}")
                .submissionAttachmentPath("/v1/submissions/{submissionId}/attachments/{attachmentId}")
                .build();

        final var envName = new EnvironmentName(environmentName);

        return ApplicationConfig.builder()
                .senderConfig(sender)
                .subscriberConfig(subscriber)
                .environments(Map.of(envName, env))
                .resourcePaths(resourcePaths)
                .activeEnvironment(envName)
                .privateSigningKeyPath("src/test/resources/private_test_signing_key.json")
                .metadataSchemaPath("src/test/resources/metadata_schema.json")
                .build();
    }
}
