package dev.fitko.fitconnect.core.validation;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.networknt.schema.JsonSchema;
import com.networknt.schema.JsonSchemaFactory;
import com.networknt.schema.SpecVersion;
import com.networknt.schema.ValidationMessage;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jose.util.Base64;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.DataIntegrityException;
import dev.fitko.fitconnect.api.exceptions.ValidationException;
import dev.fitko.fitconnect.api.services.crypto.MessageDigestService;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import dev.fitko.fitconnect.core.util.Strings;
import dev.fitko.fitconnect.jwkvalidator.JWKValidator;
import dev.fitko.fitconnect.jwkvalidator.JWKValidatorBuilder;
import dev.fitko.fitconnect.jwkvalidator.exceptions.JWKValidationException;
import dev.fitko.fitconnect.jwkvalidator.exceptions.LogLevel;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.net.InetSocketAddress;
import java.net.Proxy;
import java.security.cert.CertificateEncodingException;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import static dev.fitko.fitconnect.jwkvalidator.JWKValidator.*;

public class DefaultValidationService implements ValidationService {

    private static final Logger LOGGER = LoggerFactory.getLogger(DefaultValidationService.class);

    private final MessageDigestService messageDigestService;
    private final ApplicationConfig config;

    public DefaultValidationService(final MessageDigestService messageDigestService, final ApplicationConfig config) {
        this.messageDigestService = messageDigestService;
        this.config = config;
    }

    @Override
    public ValidationResult validatePublicKey(final RSAKey publicKey) {
        try {
            if (keyHasNoCertificateChain(publicKey)) {
                validateWithoutCertChain(publicKey);
            } else {
                validateCertChain(publicKey);
            }
        } catch (final JWKValidationException | CertificateEncodingException e) {
            return ValidationResult.error(e);
        }
        return ValidationResult.ok();
    }

    @Override
    public ValidationResult validateMetadataSchema(final String jsonSchema, final Metadata metadata) {
        try {
            final ObjectMapper mapper = new ObjectMapper();
            final String metadataJson = mapper.writeValueAsString(metadata);
            final JsonNode schemaNode = mapper.readTree(jsonSchema);
            final JsonNode inputNode = mapper.readTree(metadataJson);
            final JsonSchemaFactory factory = JsonSchemaFactory.getInstance(SpecVersion.VersionFlag.V202012);
            final JsonSchema schema = factory.getSchema(schemaNode);
            final Set<ValidationMessage> errors = schema.validate(inputNode);
            return returnValidationResult(errors);
        } catch (final JsonProcessingException e) {
            return ValidationResult.error(new ValidationException(e.getMessage(), e));
        }
    }

    @Override
    public ValidationResult validateIntegrity(final String originalHexHash, final byte[] data) {
        final byte[] originalHash = messageDigestService.fromHexString(originalHexHash);
        final boolean hashesAreNotEqual = !messageDigestService.verify(originalHash, data);
        if (hashesAreNotEqual) {
            return ValidationResult.error(new DataIntegrityException("Hash sum of transmitted data does not equal the hash of the sender."));
        }
        return ValidationResult.ok();
    }

    private ValidationResult returnValidationResult(final Set<ValidationMessage> errors) {
        if (errors.isEmpty()) {
            return ValidationResult.ok();
        }
        return ValidationResult.error(new ValidationException(errorsToSingleString(errors)));
    }

    private void validateWithX509AndProxy(final List<String> pemCerts, final RSAKey publicKey) throws JWKValidationException {
        final var proxy = new Proxy(Proxy.Type.HTTP, new InetSocketAddress(config.getHttpProxyHost(), config.getHttpProxyPort()));
        LOGGER.info("Using proxy {} for key validation", proxy);
        addLogLevel(withX5CValidation()
                .withProxy(proxy)
                .withRootCertificatesAsPEM(pemCerts))
                .validate(publicKey);
    }


    private void validateWithX509AndWithoutProxy(final List<String> pemCerts, final RSAKey publicKey) throws JWKValidationException {
        LOGGER.info("No proxy set for validation");
        addLogLevel(withX5CValidation()
                .withoutProxy()
                .withRootCertificatesAsPEM(pemCerts))
                .validate(publicKey);
    }

    private JWKValidator addLogLevel(final JWKValidatorBuilder.JWKValidatorX5CErrorHandling validator) {
        if (validateSilent()) {
            return validator.withoutThrowingExceptions().withErrorLogLevel(LogLevel.WARN).build();
        } else {
            return validator.withThrowingExceptions().withErrorLogLevel(LogLevel.ERROR).build();
        }
    }

    private void validateWithoutCertChain(final RSAKey publicKey) throws JWKValidationException {
        LOGGER.info("Validation public key without XC5 certificate chain");
        withoutX5CValidation().withErrorLogLevel(validateSilent() ? LogLevel.WARN : LogLevel.ERROR).build().validate(publicKey);
    }

    private void validateCertChain(final RSAKey publicKey) throws JWKValidationException, CertificateEncodingException {
        final List<String> pemCerts = getPemCerts(publicKey.getParsedX509CertChain());
        LOGGER.info("Validation public key with XC5 certificate chain checks");
        if (isProxySet()) {
            validateWithX509AndProxy(pemCerts, publicKey);
        } else {
            validateWithX509AndWithoutProxy(pemCerts, publicKey);
        }
    }

    private List<String> getPemCerts(final List<X509Certificate> certChain) throws CertificateEncodingException {
        final List<String> pemCerts = new ArrayList<>();
        for (final X509Certificate cert : certChain) {
            pemCerts.add(getEncodedString(cert));
        }
        return pemCerts;
    }

    private String getEncodedString(final X509Certificate x509Certificate) throws CertificateEncodingException {
        return Base64.encode(x509Certificate.getEncoded()).toString();
    }

    private boolean isProxySet() {
        return config.getHttpProxyPort() != null && Strings.isNotNullOrEmpty(config.getHttpProxyHost());
    }

    private boolean keyHasNoCertificateChain(final RSAKey publicKey) {
        return publicKey.getParsedX509CertChain() == null || publicKey.getX509CertChain().isEmpty();
    }

    private boolean validateSilent() {
        return config.getCurrentEnvironment().isSilentKeyValidation();
    }

    private String errorsToSingleString(final Set<ValidationMessage> errors) {
        return errors.stream()
                .map(ValidationMessage::getMessage)
                .collect(Collectors.joining("\n"));
    }
}
