package dev.fitko.fitconnect.core.events;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.github.tomakehurst.wiremock.http.ResponseDefinition;
import dev.fitko.fitconnect.api.domain.auth.OAuthToken;
import dev.fitko.fitconnect.api.domain.model.event.EventLog;
import dev.fitko.fitconnect.api.services.auth.OAuthService;
import dev.fitko.fitconnect.api.services.events.EventLogService;
import dev.fitko.fitconnect.core.RestEndpointBase;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mockito;
import org.springframework.http.MediaType;
import org.springframework.web.client.RestTemplate;

import java.util.List;
import java.util.UUID;

import static com.github.tomakehurst.wiremock.client.WireMock.*;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.Mockito.when;

class EventLogApiServiceTest extends RestEndpointBase {

    EventLogService underTest;
    OAuthService authServiceMock;

    @BeforeEach
    void setUp() {
        authServiceMock = Mockito.mock(OAuthService.class);
        final var config = getTestConfig( "http://localhost:" + wireMockServer.port());
        underTest = new EventLogApiService(authServiceMock, new RestTemplate(), config);
    }

    @Test
    void getEventLog() throws JsonProcessingException {
        // Given
        final var caseId = UUID.randomUUID();

        final var sampleLog = " \"https://schema.fitko.de/fit-connect/events/submit-submission\": {\n" +
                "      \"authenticationTags\": {\n" +
                "        \"metadata\": \"XFBoMYUZodetZdvTiFvSkQ\",\n" +
                "        \"data\": \"UCGiqJxhBI3IFVdPalHHvA\",\n" +
                "        \"attachments\": {\n" +
                "          \"0b799252-deb9-42b0-98d3-c50d24bbafe0\": \"rT99rwrBTbTI7IJM8fU3El\",\n" +
                "          \"25abf553-0e53-43b9-a14a-1581b32a9ee5\": \"i7226HEB7IchCxNuh7lCiu\",\n" +
                "          \"046a9fa5-bed6-494b-aab6-d41056c6db79\": \"d48LxeolRdtFF4nzQibeYO\"\n" +
                "        }\n" +
                "      }\n" +
                "    }";

        final var expectedLog = new EventLog();
        expectedLog.setEventLogs(List.of(sampleLog));

        final var authToken = new OAuthToken();
        authToken.setAccessToken("abc");
        authToken.setExpiresIn(1800);

        when(authServiceMock.getCurrentToken()).thenReturn(authToken);
        final String urlPath = "/v1/cases/" + caseId + "/events";

        wireMockServer.stubFor(
                get(urlEqualTo(urlPath))
                        .willReturn(ok()
                                .withBody(new ObjectMapper().writeValueAsString(expectedLog))
                                .withHeader("Content-Type", MediaType.APPLICATION_JSON_VALUE)
                                .withStatus(200)));

        // When
        final EventLog log = underTest.getEventLog(caseId);

        // Then
        final List<String> eventLog = log.getEventLogs();
        assertFalse(eventLog.isEmpty());
        assertTrue(eventLog.stream().findFirst().isPresent());
        assertThat(eventLog.stream().findFirst().get(), is(sampleLog));
        verify(getRequestedFor(urlEqualTo(urlPath)).withHeader("Authorization", equalTo("Bearer abc")));
    }

    @Test
    void sendEvent() {
        // Given
        final var caseId = UUID.randomUUID();
        final var serialisedSetEvent = "eyJraWQiOiJkZDA0MDllNS00MTBlLTRkOTgtODViNi1mODFhNDBiO";

        final var authToken = new OAuthToken();
        authToken.setAccessToken("abc");
        authToken.setExpiresIn(1800);

        when(authServiceMock.getCurrentToken()).thenReturn(authToken);

        final var urlPath = "/v1/cases/" + caseId + "/events";

        final ResponseDefinition response = wireMockServer.stubFor(
                post(urlEqualTo(urlPath))
                        .willReturn(ok()
                                .withBody(serialisedSetEvent)
                                .withHeader("Content-Type", MediaType.APPLICATION_JSON_VALUE)
                                .withStatus(204))).getResponse();

        // When
        underTest.sendEvent(caseId, serialisedSetEvent);

        // Then
        assertThat(response.getStatus(), is(204));
        verify(postRequestedFor(urlEqualTo(urlPath)).withHeader("Authorization", equalTo("Bearer abc")));
        verify(postRequestedFor(urlEqualTo(urlPath)).withHeader("Content-Type", equalTo("application/jose")));
    }
}