package dev.fitko.fitconnect.core.events;

import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.JWSAlgorithm;
import com.nimbusds.jose.JWSHeader;
import com.nimbusds.jose.JWSVerifier;
import com.nimbusds.jose.crypto.RSASSAVerifier;
import com.nimbusds.jose.jwk.JWK;
import com.nimbusds.jwt.JWTClaimsSet;
import com.nimbusds.jwt.SignedJWT;
import dev.fitko.fitconnect.api.exceptions.InvalidSigningKeyException;
import dev.fitko.fitconnect.api.services.events.SecurityEventService;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.text.ParseException;
import java.util.Map;
import java.util.Optional;
import java.util.UUID;

import static org.junit.jupiter.api.Assertions.*;

class SecurityEventTokenServiceTest {

    String signingKey;
    SecurityEventService underTest;

    @BeforeEach
    void setUp() throws IOException {
        final ClassLoader classLoader = getClass().getClassLoader();
        final File file = new File(classLoader.getResource("private_test_signing_key.json").getFile());
        signingKey = Files.readString(file.toPath());
        underTest = new SecurityEventTokenService(signingKey);
    }

    @Test
    void createAcceptSubmissionEvent() throws ParseException, JOSEException {

        final String event = underTest.createAcceptSubmissionEvent(UUID.randomUUID(), UUID.randomUUID(), UUID.randomUUID());

        final SignedJWT signedJWT = SignedJWT.parse(event);
        final var jwk = JWK.parse(signingKey);
        final JWTClaimsSet payload = signedJWT.getJWTClaimsSet();
        final JWSVerifier jwsVerifier = new RSASSAVerifier(jwk.toRSAKey());

        final Map<String,Object> obj = (Map<String, Object>) payload.getClaims().get("events");
        final Optional<String> first = obj.keySet().stream().findFirst();

        assertEquals(JWSAlgorithm.PS512, jwk.getAlgorithm());
        assertTrue(signedJWT.verify(jwsVerifier));
        assertEquals(SecurityEventTokenService.ACCEPT_EVENT, first.get());
    }

    @Test
    void createRejectSubmissionEvent() throws ParseException, JOSEException {

        // Given
        final String event = underTest.createRejectSubmissionEvent(UUID.randomUUID(), UUID.randomUUID(), UUID.randomUUID());

        final SignedJWT signedJWT = SignedJWT.parse(event);
        final JWSHeader header = signedJWT.getHeader();
        final var jwk = JWK.parse(signingKey);

        // Then
        assertEquals("secevent+jwt", header.getType().getType());
        assertEquals("PS512", header.getAlgorithm().getName());
        assertEquals(jwk.getKeyID(), header.getKeyID());
        assertEquals(JWSAlgorithm.PS512, jwk.getAlgorithm());

        final JWTClaimsSet payload = signedJWT.getJWTClaimsSet();
        final Map<String,Object> obj = (Map<String, Object>) payload.getClaims().get("events");
        final Optional<String> first = obj.keySet().stream().findFirst();
        final JWSVerifier jwsVerifier = new RSASSAVerifier(jwk.toRSAKey());

        assertTrue(signedJWT.verify(jwsVerifier));
        assertEquals(SecurityEventTokenService.REJECT_EVENT, first.get());
    }

    @Test
    void signingKeyCannotBeParsed(){
        assertThrows(
                InvalidSigningKeyException.class,
                () -> new SecurityEventTokenService("broken key"),
                "Expected new SecurityEventTokenService() to throw, but nothing was thrown"
        );

    }
}